{0, "Khronos", "", "Khronos"},
{1, "LunarG", "", "LunarG"},
{2, "Valve", "", "Valve"},
{3, "Codeplay", "", "Codeplay"},
{4, "NVIDIA", "", "NVIDIA"},
{5, "ARM", "", "ARM"},
{6, "Khronos", "LLVM/SPIR-V Translator", "Khronos LLVM/SPIR-V Translator"},
{7, "Khronos", "SPIR-V Tools Assembler", "Khronos SPIR-V Tools Assembler"},
{8, "Khronos", "Glslang Reference Front End", "Khronos Glslang Reference Front End"},
{9, "Qualcomm", "", "Qualcomm"},
{10, "AMD", "", "AMD"},
{11, "Intel", "", "Intel"},
{12, "Imagination", "", "Imagination"},
{13, "Google", "Shaderc over Glslang", "Google Shaderc over Glslang"},
{14, "Google", "spiregg", "Google spiregg"},
{15, "Google", "rspirv", "Google rspirv"},
{16, "X-LEGEND", "Mesa-IR/SPIR-V Translator", "X-LEGEND Mesa-IR/SPIR-V Translator"},
{17, "Khronos", "SPIR-V Tools Linker", "Khronos SPIR-V Tools Linker"},
{18, "Wine", "VKD3D Shader Compiler", "Wine VKD3D Shader Compiler"},
{19, "Tellusim", "Clay Shader Compiler", "Tellusim Clay Shader Compiler"},
{20, "W3C WebGPU Group", "WHLSL Shader Translator", "W3C WebGPU Group WHLSL Shader Translator"},
{21, "Google", "Clspv", "Google Clspv"},
{22, "LLVM", "MLIR SPIR-V Serializer", "LLVM MLIR SPIR-V Serializer"},
{23, "Google", "Tint Compiler", "Google Tint Compiler"},
{24, "Google", "ANGLE Shader Compiler", "Google ANGLE Shader Compiler"},
{25, "Netease Games", "Messiah Shader Compiler", "Netease Games Messiah Shader Compiler"},
{26, "Xenia", "Xenia Emulator Microcode Translator", "Xenia Xenia Emulator Microcode Translator"},
{27, "Embark Studios", "Rust GPU Compiler Backend", "Embark Studios Rust GPU Compiler Backend"},
{28, "gfx-rs community", "Naga", "gfx-rs community Naga"},
{29, "Mikkosoft Productions", "MSP Shader Compiler", "Mikkosoft Productions MSP Shader Compiler"},
{30, "SpvGenTwo community", "SpvGenTwo SPIR-V IR Tools", "SpvGenTwo community SpvGenTwo SPIR-V IR Tools"},
{31, "Google", "Skia SkSL", "Google Skia SkSL"},
{32, "TornadoVM", "Beehive SPIRV Toolkit", "TornadoVM Beehive SPIRV Toolkit"},
{33, "DragonJoker", "ShaderWriter", "DragonJoker ShaderWriter"},
{34, "Rayan Hatout", "SPIRVSmith", "Rayan Hatout SPIRVSmith"},
{35, "Saarland University", "Shady", "Saarland University Shady"},
{36, "Taichi Graphics", "Taichi", "Taichi Graphics Taichi"},
{37, "heroseh", "Hero C Compiler", "heroseh Hero C Compiler"},
{38, "Meta", "SparkSL", "Meta SparkSL"},
{39, "SirLynix", "Nazara ShaderLang Compiler", "SirLynix Nazara ShaderLang Compiler"},
{40, "Khronos", "Slang Compiler", "Khronos Slang Compiler"},
{41, "Zig Software Foundation", "Zig Compiler", "Zig Software Foundation Zig Compiler"},
{42, "Rendong Liang", "spq", "Rendong Liang spq"},
{43, "LLVM", "LLVM SPIR-V Backend", "LLVM LLVM SPIR-V Backend"},
{44, "Robert Konrad", "Kongruent", "Robert Konrad Kongruent"},
{45, "Kitsunebi Games", "Nuvk SPIR-V Emitter and DLSL compiler", "Kitsunebi Games Nuvk SPIR-V Emitter and DLSL compiler"},
{46, "Nintendo", "", "Nintendo"},
{47, "ARM", "", "ARM"},
{48, "Goopax", "", "Goopax"},
{49, "Icyllis Milica", "Arc3D Shader Compiler", "Icyllis Milica Arc3D Shader Compiler"},