-- This file is machine generated! any entries written below will be preserved,
-- but things like comments or anything outside the schema won't be preserved
return {
	["nop"] = 0,
	["Unrecognized"] = 1,
	["Type.BasicType.Void"] = 2,
	["Type.BasicType.Bool"] = 3,
	["Type.BasicType.Int8"] = 4,
	["Type.BasicType.Int16"] = 5,
	["Type.BasicType.Int"] = 6,
	["Type.BasicType.Int64"] = 7,
	["Type.BasicType.UInt8"] = 8,
	["Type.BasicType.UInt16"] = 9,
	["Type.BasicType.UInt"] = 10,
	["Type.BasicType.UInt64"] = 11,
	["Type.BasicType.Half"] = 12,
	["Type.BasicType.Float"] = 13,
	["Type.BasicType.Double"] = 14,
	["Type.BasicType.Char"] = 15,
	["Type.BasicType.IntPtr"] = 16,
	["Type.BasicType.UIntPtr"] = 17,
	["Type.AfterBaseType"] = 18,
	["Type.StringTypeBase.String"] = 19,
	["Type.StringTypeBase.NativeString"] = 20,
	["Type.CapabilitySet"] = 21,
	["Type.DynamicType"] = 22,
	["Type.AnyValueType"] = 23,
	["Type.RawPointerTypeBase.RawPointerType"] = 24,
	["Type.RawPointerTypeBase.RTTIPointerType"] = 25,
	["Type.RawPointerTypeBase.AfterRawPointerTypeBase"] = 26,
	["Type.ArrayTypeBase.Array"] = 27,
	["Type.ArrayTypeBase.UnsizedArray"] = 28,
	["Type.Func"] = 29,
	["Type.BasicBlock"] = 30,
	["Type.Vec"] = 31,
	["Type.Mat"] = 32,
	["Type.Conjunction"] = 33,
	["Type.Attributed"] = 34,
	["Type.Result"] = 35,
	["Type.Optional"] = 36,
	["Type.Enum"] = 37,
	["Type.DifferentialPairTypeBase.DiffPair"] = 38,
	["Type.DifferentialPairTypeBase.DiffPairUserCode"] = 39,
	["Type.DifferentialPairTypeBase.DiffRefPair"] = 40,
	["Type.BwdDiffIntermediateCtxType"] = 41,
	["Type.TensorView"] = 42,
	["Type.TorchTensor"] = 43,
	["Type.ArrayListVector"] = 44,
	["Type.Atomic"] = 45,
	["Type.BindExistentialsTypeBase.BindExistentials"] = 46,
	["Type.BindExistentialsTypeBase.BoundInterface"] = 47,
	["Type.Rate.ConstExpr"] = 48,
	["Type.Rate.SpecConst"] = 49,
	["Type.Rate.GroupShared"] = 50,
	["Type.Rate.ActualGlobalRate"] = 51,
	["Type.RateQualified"] = 52,
	["Type.Kind.Type"] = 53,
	["Type.Kind.TypeParameterPack"] = 54,
	["Type.Kind.Rate"] = 55,
	["Type.Kind.Generic"] = 56,
	["Type.PtrTypeBase.Ptr"] = 57,
	["Type.PtrTypeBase.RefParam"] = 58,
	["Type.PtrTypeBase.BorrowInParam"] = 59,
	["Type.PtrTypeBase.PseudoPtr"] = 60,
	["Type.PtrTypeBase.OutParamTypeBase.OutParam"] = 61,
	["Type.PtrTypeBase.OutParamTypeBase.BorrowInOutParam"] = 62,
	["Type.ComPtr"] = 63,
	["Type.NativePtr"] = 64,
	["Type.DescriptorHandle"] = 65,
	["Type.GLSLAtomicUint"] = 66,
	["Type.SamplerStateTypeBase.SamplerState"] = 67,
	["Type.SamplerStateTypeBase.SamplerComparisonState"] = 68,
	["Type.DefaultLayout"] = 69,
	["Type.Std140Layout"] = 70,
	["Type.Std430Layout"] = 71,
	["Type.ScalarLayout"] = 72,
	["Type.SubpassInputType"] = 73,
	["Type.TextureFootprintType"] = 74,
	["Type.TextureShape1DType"] = 75,
	["Type.TextureShape2DType"] = 76,
	["Type.TextureShape3DType"] = 77,
	["Type.TextureShapeCubeDType"] = 78,
	["Type.TextureShapeBufferType"] = 79,
	["Type.ResourceTypeBase.ResourceType.TextureTypeBase.TextureType"] = 80,
	["Type.ResourceTypeBase.ResourceType.TextureTypeBase.GLSLImageType"] = 81,
	["Type.UntypedBufferResourceType.ByteAddressBufferTypeBase.ByteAddressBuffer"] = 82,
	["Type.UntypedBufferResourceType.ByteAddressBufferTypeBase.RWByteAddressBuffer"] = 83,
	["Type.UntypedBufferResourceType.ByteAddressBufferTypeBase.RasterizerOrderedByteAddressBuffer"] = 84,
	["Type.UntypedBufferResourceType.RaytracingAccelerationStructure"] = 85,
	["Type.HLSLPatchType.InputPatch"] = 86,
	["Type.HLSLPatchType.OutputPatch"] = 87,
	["Type.GLSLInputAttachment"] = 88,
	["Type.BuiltinGenericType.HLSLStreamOutputType.PointStream"] = 89,
	["Type.BuiltinGenericType.HLSLStreamOutputType.LineStream"] = 90,
	["Type.BuiltinGenericType.HLSLStreamOutputType.TriangleStream"] = 91,
	["Type.BuiltinGenericType.MeshOutputType.Vertices"] = 92,
	["Type.BuiltinGenericType.MeshOutputType.Indices"] = 93,
	["Type.BuiltinGenericType.MeshOutputType.Primitives"] = 94,
	["Type.BuiltinGenericType.metal::mesh"] = 95,
	["Type.BuiltinGenericType.mesh_grid_properties"] = 96,
	["Type.BuiltinGenericType.HLSLStructuredBufferTypeBase.StructuredBuffer"] = 97,
	["Type.BuiltinGenericType.HLSLStructuredBufferTypeBase.RWStructuredBuffer"] = 98,
	["Type.BuiltinGenericType.HLSLStructuredBufferTypeBase.RasterizerOrderedStructuredBuffer"] = 99,
	["Type.BuiltinGenericType.HLSLStructuredBufferTypeBase.AppendStructuredBuffer"] = 100,
	["Type.BuiltinGenericType.HLSLStructuredBufferTypeBase.ConsumeStructuredBuffer"] = 101,
	["Type.BuiltinGenericType.PointerLikeType.ParameterGroupType.UniformParameterGroupType.ConstantBuffer"] = 102,
	["Type.BuiltinGenericType.PointerLikeType.ParameterGroupType.UniformParameterGroupType.TextureBuffer"] = 103,
	["Type.BuiltinGenericType.PointerLikeType.ParameterGroupType.UniformParameterGroupType.ParameterBlock"] = 104,
	["Type.BuiltinGenericType.PointerLikeType.ParameterGroupType.VaryingParameterGroupType.GLSLInputParameterGroup"] = 105,
	["Type.BuiltinGenericType.PointerLikeType.ParameterGroupType.VaryingParameterGroupType.GLSLOutputParameterGroup"] = 106,
	["Type.BuiltinGenericType.PointerLikeType.ParameterGroupType.GLSLShaderStorageBuffer"] = 107,
	["Type.RayQuery"] = 108,
	["Type.HitObject"] = 109,
	["Type.CoopVectorType"] = 110,
	["Type.CoopMatrixType"] = 111,
	["Type.TensorAddressingTensorLayoutType"] = 112,
	["Type.TensorAddressingTensorViewType"] = 113,
	["Type.MakeTensorAddressingTensorLayout"] = 114,
	["Type.MakeTensorAddressingTensorView"] = 115,
	["Type.DynamicResource"] = 116,
	["Type.struct"] = 117,
	["Type.class"] = 118,
	["Type.interface"] = 119,
	["Type.associated_type"] = 120,
	["Type.this_type"] = 121,
	["Type.rtti_type"] = 122,
	["Type.rtti_handle_type"] = 123,
	["Type.TupleTypeBase.tuple_type"] = 124,
	["Type.TupleTypeBase.TypePack"] = 125,
	["Type.TargetTuple"] = 126,
	["Type.ExpandTypeOrVal"] = 127,
	["Type.spirvLiteralType"] = 128,
	["Type.type_t"] = 129,
	["Type.WitnessTableTypeBase.witness_table_t"] = 130,
	["Type.WitnessTableTypeBase.witness_table_id_t"] = 131,
	["GlobalValueWithCode.GlobalValueWithParams.func"] = 132,
	["GlobalValueWithCode.GlobalValueWithParams.generic"] = 133,
	["GlobalValueWithCode.global_var"] = 134,
	["global_param"] = 135,
	["globalConstant"] = 136,
	["key"] = 137,
	["global_generic_param"] = 138,
	["witness_table"] = 139,
	["indexedFieldKey"] = 140,
	["thisTypeWitness"] = 141,
	["TypeEqualityWitness"] = 142,
	["global_hashed_string_literals"] = 143,
	["module"] = 144,
	["block"] = 145,
	["Constant.boolConst"] = 146,
	["Constant.integer_constant"] = 147,
	["Constant.float_constant"] = 148,
	["Constant.ptr_constant"] = 149,
	["Constant.void_constant"] = 150,
	["Constant.string_constant"] = 151,
	["Constant.blob_constant"] = 152,
	["CapabilitySet.capabilityConjunction"] = 153,
	["CapabilitySet.capabilityDisjunction"] = 154,
	["Undefined.Poison"] = 155,
	["defaultConstruct"] = 156,
	["MakeDifferentialPairBase.MakeDiffPair"] = 157,
	["MakeDifferentialPairBase.MakeDiffPairUserCode"] = 158,
	["MakeDifferentialPairBase.MakeDiffRefPair"] = 159,
	["DifferentialPairGetDifferentialBase.GetDifferential"] = 160,
	["DifferentialPairGetDifferentialBase.GetDifferentialUserCode"] = 161,
	["DifferentialPairGetDifferentialBase.GetDifferentialPtr"] = 162,
	["DifferentialPairGetPrimalBase.GetPrimal"] = 163,
	["DifferentialPairGetPrimalBase.GetPrimalUserCode"] = 164,
	["DifferentialPairGetPrimalBase.GetPrimalRef"] = 165,
	["specialize"] = 166,
	["lookupWitness"] = 167,
	["GetSequentialID"] = 168,
	["bind_global_generic_param"] = 169,
	["allocObj"] = 170,
	["globalValueRef"] = 171,
	["makeUInt64"] = 172,
	["makeVector"] = 173,
	["makeMatrix"] = 174,
	["makeMatrixFromScalar"] = 175,
	["matrixReshape"] = 176,
	["vectorReshape"] = 177,
	["makeArray"] = 178,
	["makeArrayFromElement"] = 179,
	["makeCoopVector"] = 180,
	["makeCoopVectorFromValuePack"] = 181,
	["makeStruct"] = 182,
	["makeTuple"] = 183,
	["makeTargetTuple"] = 184,
	["makeValuePack"] = 185,
	["getTargetTupleElement"] = 186,
	["getTupleElement"] = 187,
	["LoadResourceDescriptorFromHeap"] = 188,
	["LoadSamplerDescriptorFromHeap"] = 189,
	["MakeCombinedTextureSamplerFromHandle"] = 190,
	["MakeWitnessPack"] = 191,
	["Expand"] = 192,
	["Each"] = 193,
	["makeResultValue"] = 194,
	["makeResultError"] = 195,
	["isResultError"] = 196,
	["getResultError"] = 197,
	["getResultValue"] = 198,
	["getOptionalValue"] = 199,
	["optionalHasValue"] = 200,
	["makeOptionalValue"] = 201,
	["makeOptionalNone"] = 202,
	["CombinedTextureSamplerGetTexture"] = 203,
	["CombinedTextureSamplerGetSampler"] = 204,
	["call"] = 205,
	["rtti_object"] = 206,
	["alloca"] = 207,
	["updateElement"] = 208,
	["detachDerivative"] = 209,
	["bitfieldExtract"] = 210,
	["bitfieldInsert"] = 211,
	["packAnyValue"] = 212,
	["unpackAnyValue"] = 213,
	["witness_table_entry"] = 214,
	["interface_req_entry"] = 215,
	["GetWorkGroupSize"] = 216,
	["GetCurrentStage"] = 217,
	["param"] = 218,
	["field"] = 219,
	["var"] = 220,
	["load"] = 221,
	["StoreBase.store"] = 222,
	["AtomicOperation.atomicLoad"] = 223,
	["AtomicOperation.atomicStore"] = 224,
	["AtomicOperation.atomicExchange"] = 225,
	["AtomicOperation.atomicCompareExchange"] = 226,
	["AtomicOperation.atomicAdd"] = 227,
	["AtomicOperation.atomicSub"] = 228,
	["AtomicOperation.atomicAnd"] = 229,
	["AtomicOperation.atomicOr"] = 230,
	["AtomicOperation.atomicXor"] = 231,
	["AtomicOperation.atomicMin"] = 232,
	["AtomicOperation.atomicMax"] = 233,
	["AtomicOperation.atomicInc"] = 234,
	["AtomicOperation.atomicDec"] = 235,
	["LoadReverseGradient"] = 236,
	["ReverseGradientDiffPairRef"] = 237,
	["PrimalParamRef"] = 238,
	["DiffParamRef"] = 239,
	["IsDifferentialNull"] = 240,
	["get_field"] = 241,
	["get_field_addr"] = 242,
	["getElement"] = 243,
	["getElementPtr"] = 244,
	["getOffsetPtr"] = 245,
	["getAddr"] = 246,
	["castDynamicResource"] = 247,
	["getNativeStr"] = 248,
	["makeString"] = 249,
	["getNativePtr"] = 250,
	["getManagedPtrWriteRef"] = 251,
	["ManagedPtrAttach"] = 252,
	["ManagedPtrDetach"] = 253,
	["imageSubscript"] = 254,
	["imageLoad"] = 255,
	["imageStore"] = 256,
	["byteAddressBufferLoad"] = 257,
	["byteAddressBufferStore"] = 258,
	["structuredBufferLoad"] = 259,
	["structuredBufferLoadStatus"] = 260,
	["rwstructuredBufferLoad"] = 261,
	["rwstructuredBufferLoadStatus"] = 262,
	["rwstructuredBufferStore"] = 263,
	["rwstructuredBufferGetElementPtr"] = 264,
	["StructuredBufferAppend"] = 265,
	["StructuredBufferConsume"] = 266,
	["StructuredBufferGetDimensions"] = 267,
	["nonUniformResourceIndex"] = 268,
	["getNaturalStride"] = 269,
	["meshOutputRef"] = 270,
	["meshOutputSet"] = 271,
	["metalSetVertex"] = 272,
	["metalSetPrimitive"] = 273,
	["metalSetIndices"] = 274,
	["MetalCastToDepthTexture"] = 275,
	["MakeVectorFromScalar"] = 276,
	["swizzle"] = 277,
	["swizzleSet"] = 278,
	["swizzledStore"] = 279,
	["TerminatorInst.return_val"] = 280,
	["TerminatorInst.yield"] = 281,
	["TerminatorInst.UnconditionalBranch.unconditionalBranch"] = 282,
	["TerminatorInst.UnconditionalBranch.loop"] = 283,
	["TerminatorInst.ConditionalBranch.conditionalBranch"] = 284,
	["TerminatorInst.ConditionalBranch.ifElse"] = 285,
	["TerminatorInst.throw"] = 286,
	["TerminatorInst.tryCall"] = 287,
	["TerminatorInst.switch"] = 288,
	["TerminatorInst.targetSwitch"] = 289,
	["TerminatorInst.GenericAsm"] = 290,
	["TerminatorInst.UnreachableBase.missingReturn"] = 291,
	["TerminatorInst.UnreachableBase.unreachable"] = 292,
	["TerminatorInst.defer"] = 293,
	["discard"] = 294,
	["RequirePrelude"] = 295,
	["RequireTargetExtension"] = 296,
	["RequireComputeDerivative"] = 297,
	["StaticAssert"] = 298,
	["Printf"] = 299,
	["RequireMaximallyReconverges"] = 300,
	["RequireQuadDerivatives"] = 301,
	["add"] = 302,
	["sub"] = 303,
	["mul"] = 304,
	["div"] = 305,
	["irem"] = 306,
	["frem"] = 307,
	["shl"] = 308,
	["shr"] = 309,
	["cmpEQ"] = 310,
	["cmpNE"] = 311,
	["cmpGT"] = 312,
	["cmpLT"] = 313,
	["cmpGE"] = 314,
	["cmpLE"] = 315,
	["and"] = 316,
	["xor"] = 317,
	["or"] = 318,
	["logicalAnd"] = 319,
	["logicalOr"] = 320,
	["neg"] = 321,
	["not"] = 322,
	["bitnot"] = 323,
	["select"] = 324,
	["checkpointObj"] = 325,
	["loopExitValue"] = 326,
	["getStringHash"] = 327,
	["waveGetActiveMask"] = 328,
	["waveMaskBallot"] = 329,
	["waveMaskMatch"] = 330,
	["sample"] = 331,
	["sampleGrad"] = 332,
	["GroupMemoryBarrierWithGroupSync"] = 333,
	["ControlBarrier"] = 334,
	["gpuForeach"] = 335,
	["getOptiXRayPayloadPtr"] = 336,
	["getOptiXHitAttribute"] = 337,
	["getOptiXSbtDataPointer"] = 338,
	["GetVulkanRayTracingPayloadLocation"] = 339,
	["GetLegalizedSPIRVGlobalParamAddr"] = 340,
	["GetPerVertexInputArray"] = 341,
	["ResolveVaryingInputRef"] = 342,
	["ForceVarIntoStructTemporarilyBase.ForceVarIntoStructTemporarily"] = 343,
	["ForceVarIntoStructTemporarilyBase.ForceVarIntoRayPayloadStructTemporarily"] = 344,
	["MetalAtomicCast"] = 345,
	["IsTextureAccess"] = 346,
	["IsTextureScalarAccess"] = 347,
	["IsTextureArrayAccess"] = 348,
	["ExtractTextureFromTextureAccess"] = 349,
	["ExtractCoordFromTextureAccess"] = 350,
	["ExtractArrayCoordFromTextureAccess"] = 351,
	["makeArrayList"] = 352,
	["makeTensorView"] = 353,
	["allocTorchTensor"] = 354,
	["TorchGetCudaStream"] = 355,
	["TorchTensorGetView"] = 356,
	["CoopMatMapElementIFunc"] = 357,
	["allocateOpaqueHandle"] = 358,
	["BindingQuery.getRegisterIndex"] = 359,
	["BindingQuery.getRegisterSpace"] = 360,
	["Decoration.highLevelDecl"] = 361,
	["Decoration.layout"] = 362,
	["Decoration.branch"] = 363,
	["Decoration.flatten"] = 364,
	["Decoration.loopControl"] = 365,
	["Decoration.loopMaxIters"] = 366,
	["Decoration.loopExitPrimalValue"] = 367,
	["Decoration.intrinsicOp"] = 368,
	["Decoration.TargetSpecificDecoration.TargetSpecificDefinitionDecoration.target"] = 369,
	["Decoration.TargetSpecificDecoration.TargetSpecificDefinitionDecoration.targetIntrinsic"] = 370,
	["Decoration.TargetSpecificDecoration.requirePrelude"] = 371,
	["Decoration.glslOuterArray"] = 372,
	["Decoration.TargetSystemValue"] = 373,
	["Decoration.interpolationMode"] = 374,
	["Decoration.nameHint"] = 375,
	["Decoration.PhysicalType"] = 376,
	["Decoration.AlignedAddressDecoration"] = 377,
	["Decoration.BinaryInterfaceType"] = 378,
	["Decoration.transitory"] = 379,
	["Decoration.ResultWitness"] = 380,
	["Decoration.vulkanRayPayload"] = 381,
	["Decoration.vulkanRayPayloadIn"] = 382,
	["Decoration.vulkanHitAttributes"] = 383,
	["Decoration.vulkanHitObjectAttributes"] = 384,
	["Decoration.GlobalVariableShadowingGlobalParameterDecoration"] = 385,
	["Decoration.requireSPIRVVersion"] = 386,
	["Decoration.requireGLSLVersion"] = 387,
	["Decoration.requireGLSLExtension"] = 388,
	["Decoration.requireWGSLExtension"] = 389,
	["Decoration.requireCUDASMVersion"] = 390,
	["Decoration.requireCapabilityAtom"] = 391,
	["Decoration.HasExplicitHLSLBinding"] = 392,
	["Decoration.DefaultValue"] = 393,
	["Decoration.readNone"] = 394,
	["Decoration.vulkanCallablePayload"] = 395,
	["Decoration.vulkanCallablePayloadIn"] = 396,
	["Decoration.earlyDepthStencil"] = 397,
	["Decoration.precise"] = 398,
	["Decoration.public"] = 399,
	["Decoration.hlslExport"] = 400,
	["Decoration.downstreamModuleExport"] = 401,
	["Decoration.downstreamModuleImport"] = 402,
	["Decoration.patchConstantFunc"] = 403,
	["Decoration.maxTessFactor"] = 404,
	["Decoration.outputControlPoints"] = 405,
	["Decoration.outputTopology"] = 406,
	["Decoration.partitioning"] = 407,
	["Decoration.domain"] = 408,
	["Decoration.maxVertexCount"] = 409,
	["Decoration.instance"] = 410,
	["Decoration.numThreads"] = 411,
	["Decoration.fpDenormalPreserve"] = 412,
	["Decoration.fpDenormalFlushToZero"] = 413,
	["Decoration.waveSize"] = 414,
	["Decoration.availableInDownstreamIR"] = 415,
	["Decoration.GeometryInputPrimitiveTypeDecoration.pointPrimitiveType"] = 416,
	["Decoration.GeometryInputPrimitiveTypeDecoration.linePrimitiveType"] = 417,
	["Decoration.GeometryInputPrimitiveTypeDecoration.trianglePrimitiveType"] = 418,
	["Decoration.GeometryInputPrimitiveTypeDecoration.lineAdjPrimitiveType"] = 419,
	["Decoration.GeometryInputPrimitiveTypeDecoration.triangleAdjPrimitiveType"] = 420,
	["Decoration.streamOutputTypeDecoration"] = 421,
	["Decoration.entryPoint"] = 422,
	["Decoration.CudaKernel"] = 423,
	["Decoration.CudaHost"] = 424,
	["Decoration.TorchEntryPoint"] = 425,
	["Decoration.AutoPyBindCUDA"] = 426,
	["Decoration.CudaKernelFwdDiffRef"] = 427,
	["Decoration.CudaKernelBwdDiffRef"] = 428,
	["Decoration.PyBindExportFuncInfo"] = 429,
	["Decoration.PyExportDecoration"] = 430,
	["Decoration.entryPointParam"] = 431,
	["Decoration.dependsOn"] = 432,
	["Decoration.keepAlive"] = 433,
	["Decoration.noSideEffect"] = 434,
	["Decoration.bindExistentialSlots"] = 435,
	["Decoration.format"] = 436,
	["Decoration.unsafeForceInlineEarly"] = 437,
	["Decoration.ForceInline"] = 438,
	["Decoration.ForceUnroll"] = 439,
	["Decoration.SizeAndAlignment"] = 440,
	["Decoration.Offset"] = 441,
	["Decoration.LinkageDecoration.import"] = 442,
	["Decoration.LinkageDecoration.export"] = 443,
	["Decoration.TargetBuiltinVar"] = 444,
	["Decoration.UserExtern"] = 445,
	["Decoration.externCpp"] = 446,
	["Decoration.externC"] = 447,
	["Decoration.dllImport"] = 448,
	["Decoration.dllExport"] = 449,
	["Decoration.cudaDeviceExport"] = 450,
	["Decoration.COMInterface"] = 451,
	["Decoration.KnownBuiltinDecoration"] = 452,
	["Decoration.RTTI_typeSize"] = 453,
	["Decoration.AnyValueSize"] = 454,
	["Decoration.SpecializeDecoration"] = 455,
	["Decoration.SequentialIDDecoration"] = 456,
	["Decoration.DynamicDispatchWitnessDecoration"] = 457,
	["Decoration.StaticRequirementDecoration"] = 458,
	["Decoration.DispatchFuncDecoration"] = 459,
	["Decoration.TypeConstraintDecoration"] = 460,
	["Decoration.BuiltinDecoration"] = 461,
	["Decoration.requiresNVAPI"] = 462,
	["Decoration.nvapiMagic"] = 463,
	["Decoration.nvapiSlot"] = 464,
	["Decoration.noInline"] = 465,
	["Decoration.noRefInline"] = 466,
	["Decoration.DerivativeGroupQuad"] = 467,
	["Decoration.DerivativeGroupLinear"] = 468,
	["Decoration.MaximallyReconverges"] = 469,
	["Decoration.QuadDerivatives"] = 470,
	["Decoration.RequireFullQuads"] = 471,
	["Decoration.TempCallArgVar"] = 472,
	["Decoration.nonCopyable"] = 473,
	["Decoration.DynamicUniform"] = 474,
	["Decoration.alwaysFold"] = 475,
	["Decoration.output"] = 476,
	["Decoration.input"] = 477,
	["Decoration.glslLocation"] = 478,
	["Decoration.glslOffset"] = 479,
	["Decoration.vkStructOffset"] = 480,
	["Decoration.raypayload"] = 481,
	["Decoration.MeshOutputDecoration.vertices"] = 482,
	["Decoration.MeshOutputDecoration.indices"] = 483,
	["Decoration.MeshOutputDecoration.primitives"] = 484,
	["Decoration.HLSLMeshPayloadDecoration"] = 485,
	["Decoration.perprimitive"] = 486,
	["Decoration.PositionOutput"] = 487,
	["Decoration.PositionInput"] = 488,
	["Decoration.PerVertex"] = 489,
	["Decoration.StageAccessDecoration.stageReadAccess"] = 490,
	["Decoration.StageAccessDecoration.stageWriteAccess"] = 491,
	["Decoration.semantic"] = 492,
	["Decoration.constructor"] = 493,
	["Decoration.method"] = 494,
	["Decoration.packoffset"] = 495,
	["Decoration.SpecializationConstantDecoration"] = 496,
	["Decoration.UserTypeName"] = 497,
	["Decoration.CounterBuffer"] = 498,
	["Decoration.RequireSPIRVDescriptorIndexingExtensionDecoration"] = 499,
	["Decoration.spirvOpDecoration"] = 500,
	["Decoration.forwardDifferentiable"] = 501,
	["Decoration.AutoDiffOriginalValueDecoration"] = 502,
	["Decoration.AutoDiffBuiltinDecoration"] = 503,
	["Decoration.fwdDerivative"] = 504,
	["Decoration.backwardDifferentiable"] = 505,
	["Decoration.primalSubstFunc"] = 506,
	["Decoration.backwardDiffPrimalReference"] = 507,
	["Decoration.backwardDiffPropagateReference"] = 508,
	["Decoration.backwardDiffIntermediateTypeReference"] = 509,
	["Decoration.backwardDiffReference"] = 510,
	["Decoration.userDefinedBackwardDiffReference"] = 511,
	["Decoration.BackwardDerivativePrimalContextDecoration"] = 512,
	["Decoration.BackwardDerivativePrimalReturnDecoration"] = 513,
	["Decoration.PrimalContextDecoration"] = 514,
	["Decoration.loopCounterDecoration"] = 515,
	["Decoration.loopCounterUpdateDecoration"] = 516,
	["Decoration.AutodiffInstDecoration.primalInstDecoration"] = 517,
	["Decoration.AutodiffInstDecoration.diffInstDecoration"] = 518,
	["Decoration.AutodiffInstDecoration.mixedDiffInstDecoration"] = 519,
	["Decoration.AutodiffInstDecoration.RecomputeBlockDecoration"] = 520,
	["Decoration.primalValueKey"] = 521,
	["Decoration.primalElementType"] = 522,
	["Decoration.IntermediateContextFieldDifferentialTypeDecoration"] = 523,
	["Decoration.derivativeMemberDecoration"] = 524,
	["Decoration.treatAsDifferentiableDecoration"] = 525,
	["Decoration.treatCallAsDifferentiableDecoration"] = 526,
	["Decoration.differentiableCallDecoration"] = 527,
	["Decoration.optimizableTypeDecoration"] = 528,
	["Decoration.ignoreSideEffectsDecoration"] = 529,
	["Decoration.CheckpointHintDecoration.PreferCheckpointDecoration"] = 530,
	["Decoration.CheckpointHintDecoration.PreferRecomputeDecoration"] = 531,
	["Decoration.CheckpointHintDecoration.CheckpointIntermediateDecoration"] = 532,
	["Decoration.NonDynamicUniformReturnDecoration"] = 533,
	["Decoration.COMWitnessDecoration"] = 534,
	["Decoration.DifferentiableTypeDictionaryDecoration"] = 535,
	["Decoration.FloatingPointModeOverride"] = 536,
	["Decoration.spvBufferBlock"] = 537,
	["Decoration.DebugLocation"] = 538,
	["Decoration.DebugFunction"] = 539,
	["Decoration.spvBlock"] = 540,
	["Decoration.NonUniformResource"] = 541,
	["Decoration.MemoryQualifierSetDecoration"] = 542,
	["Decoration.BitFieldAccessorDecoration"] = 543,
	["makeExistential"] = 544,
	["makeExistentialWithRTTI"] = 545,
	["createExistentialObject"] = 546,
	["wrapExistential"] = 547,
	["getValueFromBoundInterface"] = 548,
	["extractExistentialValue"] = 549,
	["extractExistentialType"] = 550,
	["extractExistentialWitnessTable"] = 551,
	["isNullExistential"] = 552,
	["extractTaggedUnionTag"] = 553,
	["extractTaggedUnionPayload"] = 554,
	["BuiltinCast"] = 555,
	["bitCast"] = 556,
	["reinterpret"] = 557,
	["unmodified"] = 558,
	["outImplicitCast"] = 559,
	["inOutImplicitCast"] = 560,
	["intCast"] = 561,
	["floatCast"] = 562,
	["castIntToFloat"] = 563,
	["castFloatToInt"] = 564,
	["CastPtrToBool"] = 565,
	["CastPtrToInt"] = 566,
	["CastIntToPtr"] = 567,
	["castToVoid"] = 568,
	["PtrCast"] = 569,
	["CastEnumToInt"] = 570,
	["CastIntToEnum"] = 571,
	["EnumCast"] = 572,
	["CastUInt2ToDescriptorHandle"] = 573,
	["CastDescriptorHandleToUInt2"] = 574,
	["CastDescriptorHandleToResource"] = 575,
	["TreatAsDynamicUniform"] = 576,
	["sizeOf"] = 577,
	["alignOf"] = 578,
	["countOf"] = 579,
	["GetArrayLength"] = 580,
	["IsType"] = 581,
	["TypeEquals"] = 582,
	["IsInt"] = 583,
	["IsBool"] = 584,
	["IsFloat"] = 585,
	["IsHalf"] = 586,
	["IsUnsignedInt"] = 587,
	["IsSignedInt"] = 588,
	["IsVector"] = 589,
	["GetDynamicResourceHeap"] = 590,
	["ForwardDifferentiate"] = 591,
	["BackwardDifferentiatePrimal"] = 592,
	["BackwardDifferentiatePropagate"] = 593,
	["BackwardDifferentiate"] = 594,
	["PrimalSubstitute"] = 595,
	["DispatchKernel"] = 596,
	["CudaKernelLaunch"] = 597,
	["getEquivalentStructuredBuffer"] = 598,
	["getStructuredBufferPtr"] = 599,
	["getUntypedBufferPtr"] = 600,
	["Layout.varLayout"] = 601,
	["Layout.TypeLayout.typeLayout"] = 602,
	["Layout.TypeLayout.parameterGroupTypeLayout"] = 603,
	["Layout.TypeLayout.arrayTypeLayout"] = 604,
	["Layout.TypeLayout.streamOutputTypeLayout"] = 605,
	["Layout.TypeLayout.matrixTypeLayout"] = 606,
	["Layout.TypeLayout.existentialTypeLayout"] = 607,
	["Layout.TypeLayout.structTypeLayout"] = 608,
	["Layout.TypeLayout.tupleTypeLayout"] = 609,
	["Layout.TypeLayout.structuredBufferTypeLayout"] = 610,
	["Layout.TypeLayout.ptrTypeLayout"] = 611,
	["Layout.EntryPointLayout"] = 612,
	["Attr.pendingLayout"] = 613,
	["Attr.stage"] = 614,
	["Attr.structFieldLayout"] = 615,
	["Attr.tupleFieldLayout"] = 616,
	["Attr.caseLayout"] = 617,
	["Attr.unorm"] = 618,
	["Attr.snorm"] = 619,
	["Attr.no_diff"] = 620,
	["Attr.nonuniform"] = 621,
	["Attr.Aligned"] = 622,
	["Attr.SemanticAttr.userSemantic"] = 623,
	["Attr.SemanticAttr.systemValueSemantic"] = 624,
	["Attr.LayoutResourceInfoAttr.size"] = 625,
	["Attr.LayoutResourceInfoAttr.offset"] = 626,
	["Attr.FuncThrowType"] = 627,
	["LiveRangeMarker.liveRangeStart"] = 628,
	["LiveRangeMarker.liveRangeEnd"] = 629,
	["SpecializationDictionaryItem"] = 630,
	["GenericSpecializationDictionary"] = 631,
	["ExistentialFuncSpecializationDictionary"] = 632,
	["ExistentialTypeSpecializationDictionary"] = 633,
	["DifferentiableTypeDictionaryItem"] = 634,
	["DifferentiableTypeAnnotation"] = 635,
	["BeginFragmentShaderInterlock"] = 636,
	["EndFragmentShaderInterlock"] = 637,
	["DebugSource"] = 638,
	["DebugLine"] = 639,
	["DebugVar"] = 640,
	["DebugValue"] = 641,
	["DebugInlinedAt"] = 642,
	["DebugFunction"] = 643,
	["DebugInlinedVariable"] = 644,
	["DebugScope"] = 645,
	["DebugNoScope"] = 646,
	["DebugBuildIdentifier"] = 647,
	["EmbeddedDownstreamIR"] = 648,
	["SPIRVAsm"] = 649,
	["SPIRVAsmInst"] = 650,
	["SPIRVAsmOperand.SPIRVAsmOperandLiteral"] = 651,
	["SPIRVAsmOperand.SPIRVAsmOperandInst"] = 652,
	["SPIRVAsmOperand.SPIRVAsmOperandConvertTexel"] = 653,
	["SPIRVAsmOperand.SPIRVAsmOperandRayPayloadFromLocation"] = 654,
	["SPIRVAsmOperand.SPIRVAsmOperandRayAttributeFromLocation"] = 655,
	["SPIRVAsmOperand.SPIRVAsmOperandRayCallableFromLocation"] = 656,
	["SPIRVAsmOperand.SPIRVAsmOperandEnum"] = 657,
	["SPIRVAsmOperand.SPIRVAsmOperandBuiltinVar"] = 658,
	["SPIRVAsmOperand.SPIRVAsmOperandGLSL450Set"] = 659,
	["SPIRVAsmOperand.SPIRVAsmOperandDebugPrintfSet"] = 660,
	["SPIRVAsmOperand.SPIRVAsmOperandId"] = 661,
	["SPIRVAsmOperand.SPIRVAsmOperandResult"] = 662,
	["SPIRVAsmOperand.__truncate"] = 663,
	["SPIRVAsmOperand.__entryPoint"] = 664,
	["SPIRVAsmOperand.__sampledType"] = 665,
	["SPIRVAsmOperand.__imageType"] = 666,
	["SPIRVAsmOperand.__sampledImageType"] = 667,
	["Type.CLayout"] = 668,
	["CastUInt64ToDescriptorHandle"] = 669,
	["CastDescriptorHandleToUInt64"] = 670,
	["CastStorageToLogicalBase.CastStorageToLogical"] = 671,
	["CastStorageToLogicalBase.CastStorageToLogicalDeref"] = 672,
	["Decoration.DisableCopyEliminationDecoration"] = 673,
	["Decoration.TempCallArgImmutableVar"] = 674,
	["CastResourceToDescriptorHandle"] = 675,
	["SymbolAlias"] = 676,
	["Decoration.InParamProxyVar"] = 677,
	["Attr.MemoryScope"] = 678,
	["Undefined.LoadFromUninitializedMemory"] = 679,
	["CUDA_LDG"] = 680,
	["StoreBase.copyLogical"] = 681,
	["MakeStorageTypeLoweringConfig"] = 682,
	["Decoration.experimentalModule"] = 683,
  	["SetBase.TypeSet"] = 684,
	["SetBase.FuncSet"] = 685,
	["SetBase.WitnessTableSet"] = 686,
	["SetBase.GenericSet"] = 687,
	["Type.SetTagType"] = 689,
	["Type.TaggedUnionType"] = 690,
	["CastInterfaceToTaggedUnionPtr"] = 691,
	["GetTagForSuperSet"] = 693,
	["GetTagForMappedSet"] = 694,
	["GetTagForSpecializedSet"] = 695,
	["GetTagFromSequentialID"] = 696,
	["GetSequentialIDFromTag"] = 697,
	["GetElementFromTag"] = 698,
	["GetDispatcher"] = 699,
	["GetSpecializedDispatcher"] = 700,
	["GetTagFromTaggedUnion"] = 701,
	["GetValueFromTaggedUnion"] = 702,
	["Type.UntaggedUnionType"] = 703,
	["Type.ElementOfSetType"] = 704,
	["MakeTaggedUnion"] = 705,
	["GetTypeTagFromTaggedUnion"] = 706,
	["GetTagOfElementInSet"] = 707,
	["UnboundedTypeElement"] = 708,
	["UnboundedFuncElement"] = 709,
	["UnboundedWitnessTableElement"] = 710,
	["UnboundedGenericElement"] = 711,
	["UninitializedTypeElement"] = 712,
	["UninitializedWitnessTableElement"] = 713,
	["NoneTypeElement"] = 714,
	["NoneWitnessTableElement"] = 715,
	["GetTagForSubSet"] = 716,
	["Decoration.DisallowSpecializationWithExistentialsDecoration"] = 717,
	["makeCoopMatrixFromScalar"] = 718,
	["getOptiXPayloadRegister"] = 719,
	["setOptiXPayloadRegister"] = 720
}
