// vector-math.h
#pragma once

// We will use the GLM library for our vector math types, just for simplicity.

#include "glm/glm/glm.hpp"
#include "glm/glm/gtc/constants.hpp"
#include "glm/glm/gtc/matrix_transform.hpp"
#include "glm/glm/gtc/quaternion.hpp"

namespace gfx
{

using namespace glm;

} // namespace gfx
