// slang-ast-all.h

#pragma once

#include "slang-ast-base.h"
#include "slang-ast-decl.h"
#include "slang-ast-expr.h"
#include "slang-ast-modifier.h"
#include "slang-ast-stmt.h"
#include "slang-ast-type.h"
#include "slang-ast-val.h"
#include "slang-capability-val.h"
