#include "../core/slang-string.h"
#include "spirv/unified1/GLSL.std.450.h"
#include "spirv/unified1/spirv.h"

namespace Slang
{
bool lookupGLSLstd450(const UnownedStringSlice& str, GLSLstd450& value);
}
