#ifndef SLANG_CORE_BASIC_H
#define SLANG_CORE_BASIC_H

#include "slang-array.h"
#include "slang-common.h"
#include "slang-dictionary.h"
#include "slang-exception.h"
#include "slang-list.h"
#include "slang-math.h"
#include "slang-short-list.h"
#include "slang-smart-pointer.h"
#include "slang-string.h"

#endif
