#include "slang-emit-slang.h"

namespace Slang
{

SlangResult emitSlangDeclarationsForEntryPoints(
    CodeGenContext* codeGenContext,
    LinkedIR& linkedIR,
    String& outSlangCode)
{
    SLANG_UNUSED(codeGenContext);
    SLANG_UNUSED(linkedIR);
    SLANG_UNUSED(outSlangCode);
    return SLANG_OK;
}

} // namespace Slang
