// test-context.cpp
#include "slangi-tool.h"

#include "../../source/core/slang-exception.h"
#include "../../source/core/slang-io.h"
#include "../../source/core/slang-test-tool-util.h"

using namespace Slang;

#include "slangi-tool-impl.h"
